# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2021, Ansible Project
"""I/O helper functions."""

import os
import os.path

import aiofiles

from .. import app_context
from ..logging import log


mlog = log.fields(mod=__name__)


async def copy_file(source_path: str, dest_path: str) -> None:
    """
    Copy content from one file to another.

    :arg source_path: Source path. Must be a file.
    :arg dest_path: Destination path.
    """
    flog = mlog.fields(func='copy_file')
    flog.debug('Enter')

    lib_ctx = app_context.lib_ctx.get()
    if lib_ctx.file_check_content > 0:
        # Check whether the destination file exists and has the same content as the source file,
        # in which case we won't overwrite the destination file
        try:
            stat_d = os.stat(dest_path)
            if stat_d.st_size <= lib_ctx.file_check_content:
                stat_s = os.stat(source_path)
                if stat_d.st_size == stat_s.st_size:
                    # Read both files and compare
                    async with aiofiles.open(source_path, 'rb') as f_in:
                        content_to_copy = await f_in.read()
                    async with aiofiles.open(dest_path, 'rb') as f_in:
                        existing_content = await f_in.read()
                    if content_to_copy == existing_content:
                        flog.debug('Skipping copy, since files are identical')
                        return
                    # Since we already read the contents of the file to copy, simply write it to
                    # the destination instead of reading it again
                    async with aiofiles.open(dest_path, 'wb') as f_out:
                        await f_out.write(content_to_copy)
                    return
        except FileNotFoundError:
            # Destination (or source) file does not exist
            pass

    async with aiofiles.open(source_path, 'rb') as f_in:
        async with aiofiles.open(dest_path, 'wb') as f_out:
            # TODO: PY3.8: while chunk := await f.read(lib_ctx.chunksize)
            chunk = await f_in.read(lib_ctx.chunksize)
            while chunk:
                await f_out.write(chunk)
                chunk = await f_in.read(lib_ctx.chunksize)

    flog.debug('Leave')


async def write_file(filename: str, content: str) -> None:
    flog = mlog.fields(func='write_file')
    flog.debug('Enter')

    content_bytes = content.encode('utf-8')

    lib_ctx = app_context.lib_ctx.get()
    if lib_ctx.file_check_content > 0 and len(content_bytes) <= lib_ctx.file_check_content:
        # Check whether the destination file exists and has the same content as the one we want to
        # write, in which case we won't overwrite the file
        try:
            stat = os.stat(filename)
            if stat.st_size == len(content_bytes):
                # Read file and compare
                async with aiofiles.open(filename, 'rb') as f:
                    existing_content = await f.read()
                if existing_content == content_bytes:
                    flog.debug('Skipping write, since file already contains the exact content')
                    return
        except FileNotFoundError:
            # Destination file does not exist
            pass

    async with aiofiles.open(filename, 'wb') as f:
        await f.write(content_bytes)

    flog.debug('Leave')


async def read_file(filename: str, encoding: str = 'utf-8') -> str:
    flog = mlog.fields(func='read_file')
    flog.debug('Enter')

    async with aiofiles.open(filename, 'r', encoding=encoding) as f:
        content = await f.read()

    flog.debug('Leave')
    return content



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project
"""Entrypoint to the antsibull-docs script."""

import sh

try:
    # We fallback to /usr/bin/getfacl so we can ignore failure to import this
    import posix1e  # type: ignore[import] # pyre-ignore[21]
except ImportError:
    posix1e = None


class UnableToCheck(Exception):
    """No library or command support to check this."""


class CheckFailure(Exception):
    """We checking failed unexpectedly."""


def _get_acls(path: str) -> str:
    """Return the acls for a given file or raise an exception."""
    acls = None
    if posix1e:
        acl = posix1e.ACL(file=path)
        acls = acl.to_any_text(options=posix1e.TEXT_NUMERIC_IDS).decode('utf-8')
    else:
        try:
            # sh dynamically creates functions which map to executables
            # pyre-ignore[16]
            acls = sh.getfacl(path, '-n').stdout.decode('utf-8')  # pylint:disable=no-member
        except sh.CommandNotFound:
            pass
        except Exception as e:
            # pylint:disable-next=raise-missing-from
            raise CheckFailure(f'Error while trying to get acls for {path}: {e}')

    if not acls:
        raise UnableToCheck(f'No way to determine acls for {path}')

    return acls


def writable_via_acls(path: str, euid: int) -> bool:
    """
    Check whether acls gives someone other than the user write access to a path.

    :arg path: Pathname to check.
    :arg euid: userid to identify the user by.
    :returns: True if acls give someone other than the user path write access to the path.
    """
    # Check acls if available
    acls = _get_acls(path)

    mask_has_write = True
    principal_has_write = False

    # Strip comments and blank lines.  Everything else is an acl
    for acl in (a for a in acls.splitlines() if not a.startswith('#') and a):
        type_, principal, permissions = acl.rsplit(':', 2)
        # default acls have the same issues as acls directly on the directory
        if type_.startswith('default:'):
            dummy_, type_ = type_.split(':', 1)

        # All the safe things:

        # ACL is for the user themselves
        if type_ == 'user' and (not principal or principal == euid):
            continue

        if type_ == 'mask':
            # The mask does not allow acls to set write
            if 'w' not in permissions:
                mask_has_write = False
                break
            # No other mask values affect safety
            continue

        # The acl doesn't grant write permission
        if 'w' not in permissions:
            continue

        # Okay, the principal has write in this case
        principal_has_write = True

    if mask_has_write and principal_has_write:
        return True
    return False



# Copyright (C) 2018 Julien Hartmann
# This program is distributed under the MIT license, a copy of which you should
# have receveived along with it. If not, see <https://opensource.org/licenses/MIT>.
#
# SPDX-License-Identifier: MIT
#
# https://github.com/solid-abstractions/python/tree/master/testing-aiohttp-clients

import contextlib
import datetime
import pytest
import ssl
import tempfile


class TemporaryCertificate:
    ''' Certificate representation used by :func:`ssl_certificate` fixture '''

    def __enter__(self):
        from cryptography import x509
        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives import hashes, serialization
        from cryptography.hazmat.primitives.asymmetric import rsa

        subject = issuer = x509.Name([x509.NameAttribute(x509.oid.NameOID.COMMON_NAME, 'localhost')])

        with contextlib.ExitStack() as stack:
            key = rsa.generate_private_key(
                public_exponent=65537, key_size=2048,
                backend=default_backend()
            )

            key_file = stack.enter_context(tempfile.NamedTemporaryFile())
            key_file.write(key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            ))
            key_file.flush()


            cert = (x509.CertificateBuilder().subject_name(subject)
                                            .issuer_name(issuer)
                                            .public_key(key.public_key())
                                            .serial_number(x509.random_serial_number())
                                            .not_valid_before(datetime.datetime.utcnow())
                                            .not_valid_after(datetime.datetime.utcnow() +
                                                            datetime.timedelta(days=1))
                                            .add_extension(
                                                x509.SubjectAlternativeName([
                                                    x509.DNSName('localhost'),
                                                    x509.DNSName('127.0.0.1'),
                                                ]),
                                                critical=False,
                                            )
                                            .sign(key, hashes.SHA256(), default_backend()))

            cert_file = stack.enter_context(tempfile.NamedTemporaryFile())
            cert_file.write(cert.public_bytes(serialization.Encoding.PEM))
            cert_file.flush()

            self._key_file, self._cert_file = key_file, cert_file
            stack.pop_all()
        return self

    def __exit__(self, exc, exc_type, tb):
        self._key_file.close()
        self._cert_file.close()

    def load_verify(self, context):
        ''' Load the certificate for verification purposes.

        :param ssl.SSLContext context: a SSL context that will be associated with the server.
        '''
        context.load_verify_locations(cafile=self._cert_file.name)

    def client_context(self):
        ''' A client-side SSL context accepting the certificate, and no others '''
        context = ssl.SSLContext()
        context.verify_mode = ssl.VerifyMode.CERT_REQUIRED
        self.load_verify(context)
        return context

    def server_context(self):
        ''' A server-side SSL context using the certificate '''
        context = ssl.SSLContext()
        context.load_cert_chain(self._cert_file.name, keyfile=self._key_file.name)
        return context


@pytest.fixture(scope='session')
def ssl_certificate():
    ''' Self-signed certificate fixture, used for local server tests. '''
    with TemporaryCertificate() as certificate:
        yield certificate



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project
"""Functionality to work with a venv."""

import os
import typing as t
import venv

import sh


def get_clean_environment() -> t.Dict[str, str]:
    env = os.environ.copy()
    try:
        del env['PYTHONPATH']
    except KeyError:
        # We just wanted to make sure there was no PYTHONPATH set...
        # all python libs will come from the venv
        pass
    return env


class VenvRunner:
    """
    Makes running a command in a venv easy.

    Combines venv functionality with sh.

    .. seealso::
        * `sh <https://amoffat.github.io/sh/>`_
        * :python:mod:`venv`
    """

    def __init__(self, name: str, top_dir: str) -> None:
        """
        Create a venv.

        :arg name: Name of the venv.
        :arg top_dir: Directory the venv will be created inside of.
        """
        self.name = name
        self.top_dir = top_dir
        self.venv_dir: str = os.path.join(top_dir, name)
        venv.create(self.venv_dir, clear=True, symlinks=True, with_pip=True)
        self._python = self.get_command('python')

        # Upgrade pip to the latest version.
        # Note that cryptography stopped building manylinux1 wheels (the only ship manylinux2010) so
        # we need pip19+ in order to work now.  RHEL8 and Ubuntu 18.04 contain a pip that's older
        # than that so we must upgrade to something even if it's not latest.
        self._python('-m', 'pip', 'install', '--upgrade', 'pip')

    def get_command(self, executable_name) -> sh.Command:
        """
        Return an :sh:obj:`sh.Command` for the given program installed within the venv.

        :arg executable_name: Program to return a command for.
        :returns: An :obj:`sh.Command` that will invoke the program.
        """
        return sh.Command(os.path.join(self.venv_dir, 'bin', executable_name))

    def install_package(self, package_name: str) -> sh.RunningCommand:
        """
        Install a python package into the venv.

        :arg package_name: This can be a bare package name or a path to a file.  It's passed
            directly to :command:`pip install`.
        :returns: An :sh:obj:`sh.RunningCommand` for the pip output.
        """
        return self._python('-m', 'pip', 'install', package_name, _env=get_clean_environment())


class FakeVenvRunner:
    """
    Simply runs commands.

    .. seealso::
        * `sh <https://amoffat.github.io/sh/>`_
        * :python:mod:`venv`
    """

    @staticmethod
    def get_command(executable_name) -> sh.Command:
        """
        Return an :sh:obj:`sh.Command` for the given program installed within the venv.

        :arg executable_name: Program to return a command for.
        :returns: An :obj:`sh.Command` that will invoke the program.
        """
        return sh.Command(executable_name)

    @staticmethod
    def install_package(package_name: str) -> sh.RunningCommand:
        """
        Install a python package into the venv.

        :arg package_name: This can be a bare package name or a path to a file.  It's passed
            directly to :command:`pip install`.
        :returns: An :sh:obj:`sh.RunningCommand` for the pip output.
        """
        raise Exception('Not implemented')



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project

import argparse

import antsibull_core.app_context as ap
from antsibull_core.schemas.config import LoggingModel
from antsibull_core.utils.collections import ContextDict

#
# Context creation tests
#

def test_default():
    """Context has default values if no ocntext manager is used"""
    lib_ctx = ap.lib_ctx.get()
    assert lib_ctx.chunksize == 4096
    assert lib_ctx.process_max is None
    assert lib_ctx.thread_max == 8
    assert lib_ctx.max_retries == 10

    app_ctx = ap.app_ctx.get()
    assert app_ctx.extra == ContextDict()
    assert app_ctx.galaxy_url == 'https://galaxy.ansible.com/'
    assert isinstance(app_ctx.logging_cfg, LoggingModel)
    assert app_ctx.pypi_url == 'https://pypi.org/'


def test_create_contexts_with_cfg():
    """Test that the create_contexts function sets values from a config dict"""
    cfg = {'chunksize': 1, 'pypi_url': 'https://test.pypi.org/', 'unknown': True}
    app_ctx, lib_ctx, args, cfg = ap.create_contexts(cfg=cfg)

    assert args == argparse.Namespace()
    assert cfg == {}

    assert lib_ctx.chunksize == 1
    assert lib_ctx.process_max is None
    assert lib_ctx.thread_max == 8
    assert lib_ctx.max_retries == 10

    assert app_ctx.extra == ContextDict({'unknown': True})
    assert app_ctx.galaxy_url == 'https://galaxy.ansible.com/'
    assert isinstance(app_ctx.logging_cfg, LoggingModel)
    assert app_ctx.pypi_url == 'https://test.pypi.org/'


def test_create_contexts_with_args():
    """Test that the create_context function sets values from cli args"""
    args = {'process_max': 2, 'pypi_url': 'https://test.pypi.org/', 'unknown': True}
    args = argparse.Namespace(**args)
    app_ctx, lib_ctx, args, cfg = ap.create_contexts(args=args)

    assert args == argparse.Namespace()
    assert cfg == {}

    assert lib_ctx.chunksize == 4096
    assert lib_ctx.process_max == 2
    assert lib_ctx.thread_max == 8
    assert lib_ctx.max_retries == 10

    assert app_ctx.extra == ContextDict({'unknown': True})
    assert app_ctx.galaxy_url == 'https://galaxy.ansible.com/'
    assert isinstance(app_ctx.logging_cfg, LoggingModel)
    assert app_ctx.pypi_url == 'https://test.pypi.org/'


def test_create_contexts_with_args_and_cfg():
    """Test that args override cfg"""
    cfg = {'chunksize': 1, 'thread_max': 2, 'galaxy_url': 'https://dev.galaxy.ansible.com/',
           'pypi_url': 'https://test.pypi.org/', 'unknown': True, 'cfg': 1}
    args = {'chunksize': 3, 'pypi_url': 'https://other.pypi.org/', 'unknown': False, 'args': 2}
    args = argparse.Namespace(**args)
    app_ctx, lib_ctx, args, cfg = ap.create_contexts(args=args, cfg=cfg)

    assert args == argparse.Namespace()
    assert cfg == {}

    assert lib_ctx.chunksize == 3
    assert lib_ctx.process_max is None
    assert lib_ctx.thread_max == 2
    assert lib_ctx.max_retries == 10

    assert app_ctx.extra == ContextDict({'unknown': False, 'cfg': 1, 'args': 2})
    assert app_ctx.galaxy_url == 'https://dev.galaxy.ansible.com/'
    assert isinstance(app_ctx.logging_cfg, LoggingModel)
    assert app_ctx.pypi_url == 'https://other.pypi.org/'


def test_create_contexts_without_extra():
    """Test that use_extra=False returns unused args and cfg."""
    cfg_data = {'chunksize': 7, 'unknown': True, 'cfg': 1}
    args_data = {'thread_max': 10, 'unknown': False, 'args': 2}
    input_args = argparse.Namespace(**args_data)
    app_ctx, lib_ctx, args, cfg = ap.create_contexts(args=input_args, cfg=cfg_data,
                                                     use_extra=False)

    assert args == argparse.Namespace(unknown=False, args=2)
    assert cfg == {'unknown': True, 'cfg': 1}

    assert lib_ctx.chunksize == 7
    assert lib_ctx.process_max is None
    assert lib_ctx.thread_max == 10
    assert lib_ctx.max_retries == 10

    assert app_ctx.extra == ContextDict()
    assert app_ctx.galaxy_url == 'https://galaxy.ansible.com/'
    assert isinstance(app_ctx.logging_cfg, LoggingModel)
    assert app_ctx.pypi_url == 'https://pypi.org/'


#
# Context manager tests
#

def test_context_overrides():
    data = ap.create_contexts(cfg={'chunksize': 5, 'galaxy_url': 'https://dev.galaxy.ansible.com/'})

    with ap.app_context(data.app_ctx) as app_ctx:
        # Test that the app_context that was returned has the new values
        assert app_ctx.galaxy_url == 'https://dev.galaxy.ansible.com/'

        # Test that the context that we can retrieve has the new values too
        app_ctx = ap.app_ctx.get()
        assert app_ctx.galaxy_url == 'https://dev.galaxy.ansible.com/'

    with ap.lib_context(data.lib_ctx) as lib_ctx:
        # Test that the returned lib_ctx has the new values
        assert lib_ctx.chunksize == 5

        # Likewise for the one that we can retrieve
        lib_ctx = ap.lib_ctx.get()
        assert lib_ctx.chunksize == 5

    # Check that once we return from the context managers, the old values have been restored
    app_ctx = ap.app_ctx.get()
    assert app_ctx.extra == ContextDict()
    assert app_ctx.galaxy_url == 'https://galaxy.ansible.com/'
    assert isinstance(app_ctx.logging_cfg, LoggingModel)
    assert app_ctx.pypi_url == 'https://pypi.org/'

    lib_ctx = ap.lib_ctx.get()
    assert lib_ctx.chunksize == 4096
    assert lib_ctx.process_max is None
    assert lib_ctx.thread_max == 8
    assert lib_ctx.max_retries == 10


def test_manager_creates_new_context():

    orig_app_ctx = ap.app_ctx.get()
    with ap.app_context() as app_ctx:
        new_app_ctx = ap.app_ctx.get()
        # Test that the app_ctx that was returned is the same as the one that is now set
        assert app_ctx is new_app_ctx

        # Test that the new app_ctx is different than the old one
        assert new_app_ctx is not orig_app_ctx

        # Test that the app_ctx that was returned has the same values as the old context
        assert new_app_ctx == orig_app_ctx

    orig_lib_ctx = ap.lib_ctx.get()
    with ap.lib_context() as lib_ctx:
        new_lib_ctx = ap.lib_ctx.get()
        # Test that the lib_ctx that was returned is the same as the one that is now set
        assert lib_ctx is new_lib_ctx

        # Test that the new lib_ctx is different than the old one
        assert new_lib_ctx is not orig_lib_ctx

        # Test that the lib_ctx that was returned has the same values as the old context
        assert new_lib_ctx == orig_lib_ctx

    # Check that once we return from the context managers, the old contexts have been returned
    assert orig_app_ctx is ap.app_ctx.get()
    assert orig_lib_ctx is ap.lib_ctx.get()


def test_app_and_lib_context():
    data = ap.create_contexts(cfg={'chunksize': 5, 'galaxy_url': 'https://dev.galaxy.ansible.com/'})

    with ap.app_and_lib_context(data) as (app_ctx, lib_ctx):
        # Test that the app_context that was returned has the new values
        assert app_ctx.galaxy_url == 'https://dev.galaxy.ansible.com/'

        # Test that the context that we can retrieve has the new values too
        app_ctx = ap.app_ctx.get()
        assert app_ctx.galaxy_url == 'https://dev.galaxy.ansible.com/'

        # Test that the returned lib_ctx has the new values
        assert lib_ctx.chunksize == 5

        # Likewise for the one that we can retrieve
        lib_ctx = ap.lib_ctx.get()
        assert lib_ctx.chunksize == 5

    # Check that once we return from the context manager, the old values have been restored
    app_ctx = ap.app_ctx.get()
    assert app_ctx.extra == ContextDict()
    assert app_ctx.galaxy_url == 'https://galaxy.ansible.com/'
    assert isinstance(app_ctx.logging_cfg, LoggingModel)
    assert app_ctx.pypi_url == 'https://pypi.org/'

    lib_ctx = ap.lib_ctx.get()
    assert lib_ctx.chunksize == 4096
    assert lib_ctx.process_max is None
    assert lib_ctx.thread_max == 8
    assert lib_ctx.max_retries == 10



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project
"""General functions for working with aiohttp."""

import asyncio
import math
import random
import typing as t
import warnings

import aiohttp

from .. import app_context
from ..logging import log


mlog = log.fields(mod=__name__)


def _format_call(command: str, args: t.Tuple[t.Any, ...], kwargs: t.Mapping[str, t.Any]) -> str:
    arguments = [repr(a) for a in args] + [f'{k}={repr(v)}' for k, v in kwargs.items()]
    return f'aio_session.{command}({", ".join(arguments)})'


class RetryGetManager:
    response: t.Optional[aiohttp.ClientResponse]

    def __init__(self,
                 aio_session: 'aiohttp.client.ClientSession',
                 args: t.Tuple[t.Any, ...],
                 kwargs: t.Mapping[str, t.Any],
                 max_retries: int,
                 acceptable_error_codes: t.Iterable[int],
                 ):
        self.aio_session = aio_session
        self.args = args
        self.kwargs = kwargs
        self.max_retries = max_retries
        self.acceptable_error_codes = acceptable_error_codes
        self.call_string = _format_call('get', args, kwargs)
        self.response = None

    async def __aenter__(self) -> aiohttp.ClientResponse:
        flog = mlog.fields(func='RetryGetManager.__aenter__')
        flog.debug('Enter')

        errors = []
        for retry in range(self.max_retries):
            flog.debug('Execute {0}', self.call_string)
            wait_factor: float = 5
            try:
                response = await self.aio_session.get(*self.args, **self.kwargs, timeout=20)
                status_code = response.status
                flog.debug(f'Status code {status_code}')
                if status_code < 400 or status_code in self.acceptable_error_codes:
                    self.response = response
                    flog.debug('Leave')
                    return response
                status = str(status_code)
                response.close()
            except asyncio.TimeoutError:
                flog.trace()
                status = 'timeout'
                wait_factor = 0.5
            except Exception as error:  # pylint:disable=broad-except
                flog.trace()
                status = str(error)

            errors.append(status)
            failed = retry + 1 == self.max_retries
            warnings.warn(
                f'{self.call_string} failed with status code {status}'
                f'{", finally failed." if failed else ", retrying..."}')
            if failed:
                break

            await asyncio.sleep(math.pow(1.5, retry) * wait_factor + (0.5 + random.uniform(0, 1)))

        flog.debug('Raise error')
        raise Exception(
            f'Repeated error when calling {self.call_string}: received status codes '
            f'{", ".join(errors)}')

    async def __aexit__(self, exc_type, exc, tb) -> None:
        flog = mlog.fields(func='RetryGetManager.__aexit__')
        flog.debug('Enter')
        response = self.response
        self.response = None
        if response is not None:
            response.close()
        flog.debug('Leave')


def retry_get(aio_session: 'aiohttp.client.ClientSession',
              *args,
              acceptable_error_codes: t.Optional[t.Iterable[int]] = None,
              max_retries: t.Optional[int] = None,
              **kwargs) -> t.AsyncContextManager[aiohttp.ClientResponse]:
    # Handle default value for max_retries
    lib_ctx = app_context.lib_ctx.get()
    if max_retries is None:
        max_retries = lib_ctx.max_retries

    # Make sure max_retries is at least 1
    max_retries = max(max_retries, 1)

    return RetryGetManager(aio_session, args, kwargs, max_retries, acceptable_error_codes or ())



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project
"""General functions for working with python collections and classes for new data types."""

import typing as t
from collections.abc import Container, Mapping, Sequence, Set

from ..vendored.collections import ImmutableDict


def is_sequence(obj: t.Any, include_string: bool = False) -> bool:
    """
    Return whether ``obj`` is a :python:obj:`collections.abc.Sequence`.

    You often want to know if a variable refers to a Sequence container but treat strings as being
    scalars instead of being Sequences.  This function allows you to do that.

    :arg obj: The object to test to see if it is a Sequence.
    :kwarg include_string: When False, strings are treated as not being a Sequence.  When True,
        they are.
    :returns: True if the object is a :python:obj:`collections.abc.Sequence`, False if it is not.
        Excludes strings from being a type of Sequence if ``include_string`` is False.
    """
    if isinstance(obj, (str, bytes)) and not include_string:
        return False
    if isinstance(obj, Sequence):
        return True
    return False


def compare_all_but(dict_a: t.Mapping, dict_b: t.Mapping,
                    keys_to_ignore: t.Optional[t.Iterable] = None) -> bool:
    """
    Compare two dictionaries, with the possibility to ignore some fields.

    :arg dict_a: First dictionary to compare
    :arg dict_b: Second dictionary to compare
    :kwarg keys_to_ignore: An iterable of keys whose values in the dictionaries will not be
        compared.
    :returns: True if the dictionaries have matching values for all of the keys which were not
        ignored.  False otherwise.
    """
    if keys_to_ignore is None:
        return dict_a == dict_b

    if not isinstance(keys_to_ignore, Set):
        keys_to_ignore = frozenset(keys_to_ignore)

    length_a = len(frozenset(dict_a.keys()) - keys_to_ignore)
    length_b = len(frozenset(dict_b.keys()) - keys_to_ignore)

    if length_a != length_b:
        return False

    sentinel = object()

    for key, value in ((k, v) for k, v in dict_a.items() if k not in keys_to_ignore):
        if value != dict_b.get(key, sentinel):
            return False

    return True


def _make_contained_containers_immutable(obj: t.Union[Set, Sequence]) -> t.List:
    """
    Make contained containers into immutable containers.

    This is a helper for :func:`_make_immutable`.  It takes an iterable container and turns all
    values inside of it into an immutable container.  Be careful what containers you pass in.
    Mappings, for instance, will be processed without error but the results are likely not what you
    want because Mappings have both a key and a value.
    """
    temp_list = []
    for value in obj:
        if isinstance(value, Container):
            value = _make_immutable(value)
        temp_list.append(value)
    return temp_list


def _make_immutable(obj: t.Any) -> t.Any:
    """Recursively convert a container and objects inside of it into immutable data types."""
    if isinstance(obj, (str, bytes)):
        # Strings first because they are also sequences
        return obj

    if isinstance(obj, Mapping):
        temp_dict = {}
        for key, value in obj.items():
            if isinstance(value, Container):
                value = _make_immutable(value)
            temp_dict[key] = value
        return ImmutableDict(temp_dict)

    if isinstance(obj, Set):
        temp_sequence = _make_contained_containers_immutable(obj)
        return frozenset(temp_sequence)

    if isinstance(obj, Sequence):
        temp_sequence = _make_contained_containers_immutable(obj)
        return tuple(temp_sequence)

    return obj


class ContextDict(ImmutableDict):
    def __init__(self, *args, **kwargs) -> None:
        if not kwargs and len(args) == 1 and isinstance(args[0], Mapping):
            # Avoid making an intermediate dict if we were only passed a dict to initialize with
            tmp_dict = args[0]
        else:
            # Otherwise we need the dict constructor to initialize a new dict for us
            tmp_dict = dict(*args, **kwargs)

        toplevel = {}
        for key, value in tmp_dict.items():
            toplevel[key] = _make_immutable(value)
        super().__init__(toplevel)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate_and_convert

    @classmethod
    def validate_and_convert(cls, value: t.Mapping) -> 'ContextDict':
        if isinstance(value, ContextDict):
            # optimization.  If it's already an ImmutableContext, we don't need to recursively
            # convert things to immutable again.
            return value

        # Typically this will convert from a dict to an ImmutableContext
        return cls(value)



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2021, Ansible Project
"""Schemas for config files."""

import os.path
import typing as t

import pydantic as p
import twiggy.formats  # type: ignore[import]
import twiggy.outputs  # type: ignore[import]

from .validators import convert_bool, convert_none, convert_path


#: Valid choices for a logging level field
LEVEL_CHOICES_F = p.Field(..., regex='^(CRITICAL|ERROR|WARNING|NOTICE|INFO|DEBUG|DISABLED)$')

#: Valid choices for a logging level field
DOC_PARSING_BACKEND_CHOICES_F = p.Field(
    'ansible-internal', regex='^(auto|ansible-doc|ansible-core-2.13|ansible-internal)$')

#: Valid choice of the logging version field
VERSION_CHOICES_F = p.Field(..., regex=r'1\.0')


#
# Configuration file schema
#

class BaseModel(p.BaseModel):
    class Config:
        allow_mutation = False
        extra = p.Extra.forbid
        validate_all = True


# pyre-ignore[13]: BaseModel initializes attributes when data is loaded
class LogFiltersModel(BaseModel):
    filter: t.Union[str, t.Callable]
    args: t.Sequence[t.Any] = []
    kwargs: t.Mapping[str, t.Any] = {}


# pyre-ignore[13]: BaseModel initializes attributes when data is loaded
class LogEmitterModel(BaseModel):
    output_name: str
    level: str = LEVEL_CHOICES_F
    filters: t.List[LogFiltersModel] = []


# pyre-ignore[13]: BaseModel initializes attributes when data is loaded
class LogOutputModel(BaseModel):
    output: t.Union[str, t.Callable]
    args: t.Sequence[t.Any] = []
    format: t.Union[str, t.Callable] = twiggy.formats.line_format
    kwargs: t.Mapping[str, t.Any] = {}

    @p.validator('args')
    # pylint:disable=no-self-argument,no-self-use
    def expand_home_dir_args(cls, args_field: t.MutableSequence,
                             values: t.Mapping) -> t.MutableSequence:
        """Expand tilde in the arguments of specific outputs."""
        if values['output'] in ('twiggy.outputs.FileOutput', twiggy.outputs.FileOutput):
            if args_field:
                args_field[0] = os.path.expanduser(args_field[0])
        return args_field

    @p.validator('kwargs')
    # pylint:disable=no-self-argument,no-self-use
    def expand_home_dir_kwargs(cls, kwargs_field: t.MutableMapping,
                               values: t.Mapping) -> t.MutableMapping:
        """Expand tilde in the keyword arguments of specific outputs."""
        if values['output'] in ('twiggy.outputs.FileOutput', twiggy.outputs.FileOutput):
            if 'name' in kwargs_field:
                kwargs_field['name'] = os.path.expanduser(kwargs_field['name'])
        return kwargs_field


class LoggingModel(BaseModel):
    emitters: t.Optional[t.Dict[str, LogEmitterModel]] = {}
    incremental: bool = False
    outputs: t.Optional[t.Dict[str, LogOutputModel]] = {}
    version: str = VERSION_CHOICES_F


#: Default logging configuration
DEFAULT_LOGGING_CONFIG = LoggingModel.parse_obj(
    {'version': '1.0',
     'outputs': {
         'logfile': {
             'output': 'twiggy.outputs.FileOutput',
             'args': [
                 '~/antsibull.log'
             ]
         },
         'stderr': {
             'output': 'twiggy.outputs.StreamOutput',
             'format': 'twiggy.formats.shell_format'
         },
     },
     'emitters': {
         'all': {
             'level': 'INFO',
             'output_name': 'logfile',
             'filters': []
         },
         'problems': {
             'level': 'WARNING',
             'output_name': 'stderr',
             'filters': []
         },
     }
     })


# This class is no longer needed, it will eventually be removed
class ConfigModel(BaseModel):
    # pyre-ignore[8]: https://github.com/samuelcolvin/pydantic/issues/1684
    ansible_base_url: p.HttpUrl = 'https://github.com/ansible/ansible'  # type: ignore[assignment]
    breadcrumbs: p.StrictBool = True
    chunksize: int = 4096
    doc_parsing_backend: str = DOC_PARSING_BACKEND_CHOICES_F
    # pyre-ignore[8]: https://github.com/samuelcolvin/pydantic/issues/1684
    galaxy_url: p.HttpUrl = 'https://galaxy.ansible.com/'  # type: ignore[assignment]
    indexes: p.StrictBool = True
    logging_cfg: LoggingModel = DEFAULT_LOGGING_CONFIG
    max_retries: int = 10
    process_max: t.Optional[int] = None
    # pyre-ignore[8]: https://github.com/samuelcolvin/pydantic/issues/1684
    pypi_url: p.HttpUrl = 'https://pypi.org/'  # type: ignore[assignment]
    use_html_blobs: p.StrictBool = False
    thread_max: int = 8
    file_check_content: int = 262144
    collection_cache: t.Optional[str] = None

    _convert_nones = p.validator('process_max', pre=True, allow_reuse=True)(convert_none)
    _convert_bools = p.validator('breadcrumbs', 'indexes', 'use_html_blobs',
                                 pre=True, allow_reuse=True)(convert_bool)
    _convert_paths = p.validator('collection_cache',
                                 pre=True, allow_reuse=True)(convert_path)



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project
"""Compat for older versions of Python."""

import asyncio
import sys


def _dummy():
    pass


if sys.version_info < (3, 7):
    #: In Python3.7+, :python:func:`asyncio.get_running_loop` is the preferred way to get the loop.
    #: It gets the loop currently running.  When that's not available, we can use
    #: :python:func:`asyncio.get_event_loop` which gets the default loop.  These are close to the
    #: same in our code since we use the default loop.  Just remember not to run it outside of an
    #: event loop.
    best_get_loop = asyncio.get_event_loop

    _loop = asyncio.get_event_loop()
    #: In Python3.7+, :python:func:`asyncio.run` is the preferred way to enter an asyncio event loop
    #: and start to execute asynchronous code.  In earlier Pythons,
    #: :python:func:`run_until_complete` can be used in almost the same manner since we are using
    #: the default event loop.
    asyncio_run = _loop.run_until_complete

    #: In Python3.7+, :python:func:`asyncio.create_task` is the preferred way to schedule additional
    #: coroutines for execution.  In earlier Pythons, :python:func:`asyncio.ensure_future` can be
    #: used, it's just not as readable.
    create_task = asyncio.ensure_future
else:
    best_get_loop = asyncio.get_running_loop
    asyncio_run = asyncio.run
    create_task = asyncio.create_task

if sys.version_info < (3, 8):
    #: In Python3.8+, :python:module:`importlib.metadata` exists to get information about installed
    #: python packages.  In earlier versions of Python, we can use the importlib_metadata backport
    #: from pypi.
    import importlib_metadata as metadata
else:
    from importlib import metadata

if sys.version_info < (3, 9):
    #: In Python3.9+, argparse.BooleanOptionalAction gives us a simple way to add
    #: --feature/--no-feature command line switches.  In earluer versions, we use
    #: code that we've copied from the upstream code.
    from .vendored._argparse_booleanoptionalaction import BooleanOptionalAction
else:
    # As long as https://bugs.python.org/issue46080 is not fixed, still use vendored version so
    # that users can actually use `--help` for affected subcommands. Once that is fixed and
    # included in released Python versions, we can adjust the sys.version_info check above.
    #
    #     from argparse import BooleanOptionalAction
    from .vendored._argparse_booleanoptionalaction import BooleanOptionalAction

__all__ = ('BooleanOptionalAction', 'asyncio_run', 'best_get_loop', 'create_task', 'metadata')



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project

from packaging.version import Version as PypiVer
from semantic_version import Version as SemVer

import pytest

from antsibull_core.dependency_files import BuildFile


SIMPLE_TEST_FILE = """_ansible_version: 4
_ansible_base_version: 2.11.0rc1
community.general: >=1.0.0,<2.0.0
community.routeros: >=2.0.0-a2,<3.0.0
"""

SIMPLE_TEST_DEPS = {'community.general': SemVer('1.0.0'),
                    'community.routeros': SemVer('2.0.0-a2'),
                    }

SIMPLE_TEST_FILE_2 = """_ansible_version: 6
_ansible_core_version: 2.13.0rc1
community.general: >=1.0.0,<2.0.0
community.routeros: >=2.0.0-a2,<3.0.0
"""

SIMPLE_TEST_DEPS_2 = {'community.general': SemVer('1.0.0'),
                    'community.routeros': SemVer('2.0.0-a2'),
                    }

@pytest.mark.parametrize('ansible_ver, core_ver, dependencies, file_contents', (
    ('4.0.0', '2.11.0rc1', SIMPLE_TEST_DEPS, SIMPLE_TEST_FILE),
    ('6.0.0', '2.13.0rc1', SIMPLE_TEST_DEPS_2, SIMPLE_TEST_FILE_2),
))
def test_build_file_write(tmpdir, ansible_ver, core_ver, dependencies, file_contents):
    filename = tmpdir / 'test.build'
    bf = BuildFile(filename)
    bf.write(PypiVer(ansible_ver), core_ver, dependencies)

    with open(filename) as f:
        assert f.read() == file_contents



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project
"""Functions to handle config files."""

import os.path
import typing as t

import perky  # type: ignore[import]
import pydantic as p

from .logging import log
from .schemas.config import ConfigModel
from .schemas.context import AppContext, LibContext


mlog = log.fields(mod=__name__)

#: System config file location.
SYSTEM_CONFIG_FILE = '/etc/antsibull.cfg'

#: Per-user config file location.
USER_CONFIG_FILE = '~/.antsibull.cfg'


class ConfigError(Exception):
    pass


def find_config_files(conf_files: t.Iterable[str]) -> t.List[str]:
    """
    Find all config files that exist.

    :arg conf_files: An iterable of config filenames to search for.
    :returns: A List of filenames which actually existed on the system.
    """
    flog = mlog.fields(func='find_config_file')
    flog.fields(conf_files=conf_files).debug('Enter')

    paths = [os.path.abspath(p) for p in conf_files]
    flog.fields(paths=paths).info('Paths to check')

    config_files = []
    for conf_path in paths:
        if os.path.exists(conf_path):
            config_files.append(conf_path)
    flog.fields(paths=config_files).info('Paths found')

    flog.debug('Leave')
    return config_files


def read_config(filename: str) -> ConfigModel:
    """
    Parse a config file and return the data from it.

    :arg filename: The filename of the config file to parse.
    :returns: A ConfigModel model containing the config data.
    """
    flog = mlog.fields(func='read_config')
    flog.debug('Enter')

    filename = os.path.abspath(filename)

    flog.fields(filename=filename).info('loading config file')
    raw_config_data = perky.load(filename)
    flog.debug('Validatinging the config file data')
    # Note: We parse the object but discard the model because we want to validate the config but let
    # the context handle all setting of defaults
    ConfigModel.parse_obj(raw_config_data)

    flog.debug('Leave')
    return raw_config_data


def validate_config(config: t.Mapping, filenames: t.List[str],
                    app_context_model: t.Type[AppContext]) -> None:
    """
    Validate configuration.

    Given the configuration loaded from one or more configuration files and the app context model,
    splits up the config into lib context and app context part and validates both parts with the
    given model. Raises a :obj:`ConfigError` if validation fails.
    """
    lib_fields = set(LibContext.__fields__)
    lib = {}
    app = {}
    for key, value in config.items():
        if key in lib_fields:
            lib[key] = value
        else:
            app[key] = value
    # Note: We parse the object but discard the model because we want to validate the config but let
    # the context handle all setting of defaults
    try:
        LibContext.parse_obj(lib)
        app_context_model.parse_obj(app)
    except p.ValidationError as exc:
        raise ConfigError(
            f"Error while parsing configuration from {', '.join(filenames)}:\n{exc}") from exc


def _load_config_file(filename: str) -> t.Mapping:
    """
    Load configuration from one file and return the raw data.
    """
    try:
        return perky.load(filename)
    except OSError as exc:
        raise ConfigError(
            f"Error while loading configuration from {filename}: {exc}") from exc
    except perky.PerkyFormatError as exc:
        raise ConfigError(
            f"Error while parsing configuration from {filename}:\n{exc}") from exc


def load_config(conf_files: t.Union[t.Iterable[str], str, None] = None,
                app_context_model: t.Type[AppContext] = AppContext) -> t.Dict:
    """
    Load configuration.

    Load configuration from all found conf files.  The default configuration is loaded
    followed by a system-wide location, user-location, and then any files specified in
    the ``conf_files`` parameter.  Toplevel keys in later files will overwrite earlier
    those same keys in earlier files.

    :arg conf_files: An iterable of conf_files to load configuration information from.
    :kwarg app_context_model: The model to use for the app context. Must be derived from
        :obj:`AppContext`. If not provided, will use :obj:`AppContext` itself.
    :returns: A dict containing the configuration.
    """
    flog = mlog.fields(func='load_config')
    flog.debug('Enter')

    if isinstance(conf_files, str):
        conf_files = (conf_files,)
    elif conf_files is None:
        conf_files = ()

    implicit_files = find_config_files((SYSTEM_CONFIG_FILE, os.path.expanduser(USER_CONFIG_FILE)))
    explicit_files = find_config_files(conf_files)

    flog.fields(implicit_files=implicit_files,
                explicit_files=explicit_files).debug('found config files')

    flog.debug('loading implicit config files')
    cfg: t.Dict = {}
    for filename in implicit_files:
        cfg.update(_load_config_file(filename))

    flog.debug('validating implicit configuration')
    validate_config(cfg, implicit_files, AppContext)

    flog.debug('loading explicit config files')
    for filename in explicit_files:
        cfg.update(_load_config_file(filename))

    flog.debug('validating combined configuration')
    validate_config(cfg, implicit_files + explicit_files, app_context_model)

    flog.fields(config=cfg).debug('Leave')
    return cfg



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2021, Ansible Project

"""
YAML handling.
"""

import typing as t

import yaml

_SafeLoader: t.Any
_SafeDumper: t.Any
try:
    # use C version if possible for speedup
    from yaml import CSafeLoader as _SafeLoader
    from yaml import CSafeDumper as _SafeDumper
except ImportError:
    from yaml import SafeLoader as _SafeLoader
    from yaml import SafeDumper as _SafeDumper


def load_yaml_bytes(data: bytes) -> t.Any:
    """
    Load and parse YAML from given bytes.
    """
    return yaml.load(data, Loader=_SafeLoader)


def load_yaml_file(path: str) -> t.Any:
    """
    Load and parse YAML file ``path``.
    """
    with open(path, 'rb') as stream:
        return yaml.load(stream, Loader=_SafeLoader)


def store_yaml_file(path: str, content: t.Any) -> None:
    """
    Store ``content`` as YAML file under ``path``.
    """
    with open(path, 'wb') as stream:
        dumper = _SafeDumper
        dumper.ignore_aliases = lambda *args: True
        yaml.dump(content, stream, default_flow_style=False, encoding='utf-8', Dumper=dumper)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['antsibull_core',
 'antsibull_core.schemas',
 'antsibull_core.utils',
 'antsibull_core.vendored',
 'tests',
 'tests.functional',
 'tests.units']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML',
 'aiofiles',
 'aiohttp>=3.0.0',
 'packaging>=20.0',
 'perky',
 'pydantic',
 'semantic_version',
 'sh',
 'twiggy>=0.5.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata'],
 ':python_version >= "3.6" and python_version < "3.7"': ['aiocontextvars']}

setup_kwargs = {
    'name': 'antsibull-core',
    'version': '1.3.0',
    'description': 'Tools for building the Ansible Distribution',
    'long_description': '<!--\nCopyright (c) Ansible Project\nGNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)\nSPDX-License-Identifier: GPL-3.0-or-later\n-->\n\n# antsibull-core -- Library for Ansible Build Scripts\n[![Python linting badge](https://github.com/ansible-community/antsibull-core/workflows/Python%20linting/badge.svg?event=push&branch=main)](https://github.com/ansible-community/antsibull-core/actions?query=workflow%3A%22Python+linting%22+branch%3Amain)\n[![Python testing badge](https://github.com/ansible-community/antsibull-core/workflows/Python%20testing/badge.svg?event=push&branch=main)](https://github.com/ansible-community/antsibull-core/actions?query=workflow%3A%22Python+testing%22+branch%3Amain)\n[![Codecov badge](https://img.shields.io/codecov/c/github/ansible-community/antsibull-core)](https://codecov.io/gh/ansible-community/antsibull-core)\n\nLibrary needed for tooling for building various things related to Ansible.\n\nYou can find a list of changes in [the antsibull-core changelog](./CHANGELOG.rst).\n\nUnless otherwise noted in the code, it is licensed under the terms of the GNU\nGeneral Public License v3 or, at your option, later.\n\nantsibull-core is covered by the [Ansible Code of Conduct](https://docs.ansible.com/ansible/latest/community/code_of_conduct.html).\n\n## Versioning and compatibility\n\nFrom version 1.0.0 on, antsibull-core sticks to semantic versioning and aims at providing no backwards compatibility breaking changes during a major release cycle. We might make exceptions from this in case of security fixes for vulnerabilities that are severe enough.\n\n## Creating a new release:\n\nIf you want to create a new release::\n\n    vim pyproject.toml  # Make sure version number is correct\n    vim changelogs/fragment/$VERSION_NUMBER.yml  # create \'release_summary:\' fragment\n    antsibull-changelog release --version $VERSION_NUMBER\n    git add CHANGELOG.rst changelogs\n    git commit -m "Release $VERSION_NUMBER."\n    poetry build\n    poetry publish  # Uploads to pypi.  Be sure you really want to do this\n\n    git tag $VERSION_NUMBER\n    git push --tags\n    vim pyproject.toml  # Bump the version number to X.Y.Z.post0\n    git commit -m \'Update the version number for the next release\' pyproject.toml\n    git push\n\n## License\n\nUnless otherwise noted in the code, it is licensed under the terms of the GNU\nGeneral Public License v3 or, at your option, later. See\n[LICENSES/GPL-3.0-or-later.txt](https://github.com/ansible-community/antsibull-changelog/tree/main/LICENSE)\nfor a copy of the license.\n\nParts of the code are vendored from other sources and are licensed under other licenses:\n1. `src/antsibull_core/vendored/collections.py` and `src/antsibull_core/vendored/json_utils.py` are licensed under the terms of the BSD 2-Clause license. See [LICENSES/BSD-2-Clause.txt](https://github.com/ansible-community/antsibull-changelog/tree/main/LICENSES/BSD-2-Clause.txt) for a copy of the license.\n2. `tests/functional/aiohttp_utils.py` and `tests/functional/certificate_utils.py` are licensed under the terms of the MIT license. See [LICENSES/MIT.txt](https://github.com/ansible-community/antsibull-changelog/tree/main/LICENSES/MIT.txt) for a copy of the license.\n3. `src/antsibull_core/vendored/_argparse_booleanoptionalaction.py` is licensed under the terms of the Python Software Foundation license version 2. See [LICENSES/PSF-2.0.txt](https://github.com/ansible-community/antsibull-changelog/tree/main/LICENSES/PSF-2.0.txt) for a copy of the license.\n\nThe repository follows the [REUSE Specification](https://reuse.software/spec/) for declaring copyright and\nlicensing information. The only exception are changelog fragments in ``changelog/fragments/``.\n',
    'author': 'Toshio Kuratomi',
    'author_email': 'a.badger@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ansible-community/antsibull-core',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2020, Ansible Project

"""
Persist collection infornation used to build Ansible.

Build dependency files list the dependencies of an Ansible release along with the versions that are
compatible with that release.

When we initially build an Ansible major release, we'll use certain versions of collections.  We
don't want to install backwards incompatible collections until the next major Ansible release.
"""

from typing import TYPE_CHECKING, Dict, List, Mapping, NamedTuple, Optional, Union

from packaging.version import Version as PypiVer

if TYPE_CHECKING:
    from semantic_version import Version as SemVer


class DependencyFileData(NamedTuple):
    ansible_version: str
    ansible_core_version: str
    deps: Dict[str, str]


class InvalidFileFormat(Exception):
    pass


def parse_pieces_file(pieces_file: str) -> List[str]:
    with open(pieces_file, 'rb') as f:
        contents = f.read()

    decoded_contents = contents.decode('utf-8')
    # One collection per line, ignoring comments and empty lines
    # TODO: PY3.8:
    # collections = [c for line in decoded_contents.split('\n')
    #                if (c := line.strip()) and not c.startswith('#')]
    collections = [line.strip() for line in decoded_contents.split('\n')
                   if line.strip() and not line.strip().startswith('#')]
    return collections


def _parse_name_version_spec_file(filename: str) -> DependencyFileData:
    deps: Dict[str, str] = {}
    ansible_core_version: Optional[str] = None
    ansible_version: Optional[str] = None

    for line in parse_pieces_file(filename):
        record = [entry.strip() for entry in line.split(':', 1)]

        if record[0] in ('_ansible_version', '_acd_version'):
            if ansible_version is not None:
                raise InvalidFileFormat(f'{filename} specified _ansible_version/_acd_version'
                                        ' more than once')
            ansible_version = record[1]
            continue

        if record[0] in ('_ansible_base_version', '_ansible_core_version'):
            if ansible_core_version is not None:
                raise InvalidFileFormat(
                    f'{filename} specified _ansible_base_version/_ansible_core_version more than'
                    ' once')
            ansible_core_version = record[1]
            continue

        deps[record[0]] = record[1]

    if ansible_core_version is None:
        raise InvalidFileFormat(f'{filename} was invalid.  It did not contain'
                                ' the required ansible_core_version field')
    if ansible_version is None:
        raise InvalidFileFormat(f'{filename} was invalid.  It did not contain'
                                ' the required ansible_version field')

    return DependencyFileData(ansible_version, ansible_core_version, deps)


class DepsFile:
    """
    File containing the collections which are part of an Ansible release.

    A DepsFile contains a sequence of lines with a collection name, ": ", and then an exact
    version of the collection.  It tracks the exact collection-versions which were installed
    with a particular ansible version.

    The deps file has two special lines which are not collections.  They are::

        _ansible_version: X1.Y1.Z1
        _ansible_core_version: X2.Y2.Z2

    (Instead of _ansible_core_version, _ansible_base_version can also be used.)

    These are, respectively, the ansible version that was built and the ansible-core version which
    it was built against.  Note that the ansible release will depend on a compatible version of that
    ansible-core version, not an exact dependency on that precise version.
    """

    def __init__(self, deps_file: str) -> None:
        """
        Create a :mod:`DepsFile`.

        :arg deps_file: filename of the `DepsFile`.
        """
        self.filename: str = deps_file

    def parse(self) -> DependencyFileData:
        """Parse the deps from a dependency file."""
        return _parse_name_version_spec_file(self.filename)

    def write(self, ansible_version: Union[str, 'PypiVer'],
              ansible_core_version: Union[str, 'PypiVer'],
              included_versions: Union[Mapping[str, str], Mapping[str, 'SemVer']],
              python_requires: Optional[str] = None) -> None:
        """
        Write a list of all the dependent collections included in this Ansible release.

        :arg ansible_version: The version of Ansible that is being recorded.
        :arg ansible_core_version: The version of Ansible-core that will be depended on.
        :arg included_versions: Dictionary mapping collection names to the version range in this
            version of Ansible.
        :arg python_requires: A python_requires string. Will be stored as ``_python``.

        WARNING: This function will no longer accept version objects in the ansible_core_version
                 and included_versions parameters, and will require a PypiVer object in the
                 ansible_version parameter in antsibull-core 2.0.0.
        """
        records = []
        for dep, version in included_versions.items():
            records.append(f'{dep}: {version}')
        records.sort()

        if not isinstance(ansible_version, PypiVer):
            ansible_version = PypiVer(ansible_version)

        with open(self.filename, 'w', encoding='utf-8') as f:
            f.write(f'_ansible_version: {ansible_version}\n')
            if ansible_version.major > 5:
                f.write(f'_ansible_core_version: {ansible_core_version}\n')
            else:
                f.write(f'_ansible_base_version: {ansible_core_version}\n')
            if python_requires is not None:
                f.write(f'_python: {python_requires}\n')
            f.write('\n'.join(records))
            f.write('\n')


class BuildFile:
    def __init__(self, build_file: str) -> None:
        self.filename: str = build_file

    def parse(self) -> DependencyFileData:
        """Parse the build from a dependency file."""
        return _parse_name_version_spec_file(self.filename)

    def write(self, ansible_version: 'PypiVer', ansible_core_version: str,
              dependencies: Mapping[str, 'SemVer'],
              python_requires: Optional[str] = None) -> None:
        """
        Write a build dependency file.

        A build dependency file records the collections that went into a given Ansible release along
        with the range of versions that are allowed for that collection.  The range is meant to
        define the set of collection versions that are compatible with what was present in the
        collection as of the first beta release, when we feature freeze the collections.

        :arg ansible_version: The version of Ansible that is being recorded.
        :arg ansible_core_version: The version of Ansible-core that will be depended on.
        :arg dependencies: Dictionary with keys of collection names and values of the latest
            versions of those collections.
        :arg python_requires: A python_requires string. Will be stored as ``_python``.
        """
        records = []
        for dep, version in dependencies.items():
            if version.prerelease and version.patch == 0:
                # Prereleases (ex: 2.0.0-b1, 2.1.0-a1) are a special case because they sort before
                # the major.minor that they are for.
                records.append(f'{dep}: >={version},<{version.major + 1}.0.0')
            else:
                records.append(f'{dep}: >={version.major}.{version.minor}.0,'
                               f'<{version.next_major()}')
        records.sort()

        with open(self.filename, 'w', encoding='utf-8') as f:
            if ansible_version.major > 2:
                # Ansible 3.0.0 and newer use semver, so we only need the major version
                f.write(f'_ansible_version: {ansible_version.major}\n')
            else:
                f.write(f'_ansible_version: {ansible_version.major}.{ansible_version.minor}\n')
            if ansible_version.major > 5:
                f.write(f'_ansible_core_version: {ansible_core_version}\n')
            else:
                f.write(f'_ansible_base_version: {ansible_core_version}\n')
            if python_requires is not None:
                f.write(f'_python: {python_requires}\n')
            f.write('\n'.join(records))
            f.write('\n')



# BSD 2-clause license (see LICENSES/BSD-2-Clause.txt)
# SPDX-FileCopyrightText: Ansible Project
# SPDX-License-Identifier: BSD-2-Clause

# pylint:disable=missing-module-docstring,no-else-break,redundant-u-string-prefix
# pylint:disable=undefined-loop-variable


# NB: a copy of this function exists in ../../modules/core/async_wrapper.py. Ensure any
# changes are propagated there.
def _filter_non_json_lines(data):
    '''
    Used to filter unrelated output around module JSON output, like messages from
    tcagetattr, or where dropbear spews MOTD on every single command (which is nuts).

    Filters leading lines before first line-starting occurrence of '{' or '[', and filter all
    trailing lines after matching close character (working from the bottom of output).
    '''
    warnings = []

    # Filter initial junk
    lines = data.splitlines()

    for start, line in enumerate(lines):
        line = line.strip()
        if line.startswith(u'{'):
            endchar = u'}'
            break
        elif line.startswith(u'['):
            endchar = u']'
            break
    else:
        raise ValueError('No start of json char found')

    # Filter trailing junk
    lines = lines[start:]

    for reverse_end_offset, line in enumerate(reversed(lines)):
        if line.strip().endswith(endchar):
            break
    else:
        raise ValueError('No end of json char found')

    if reverse_end_offset > 0:
        # Trailing junk is uncommon and can point to things the user might
        # want to change.  So print a warning if we find any
        trailing_junk = lines[len(lines) - reverse_end_offset:]
        for line in trailing_junk:
            if line.strip():
                warnings.append('Module invocation had junk after the JSON data: %s'
                                % '\n'.join(trailing_junk))
                break

    lines = lines[:(len(lines) - reverse_end_offset)]

    return ('\n'.join(lines), warnings)
